import logging
import re

from kinto.core.errors import raise_invalid
from kinto.core.events import AfterResourceChanged, ResourceChanged
from kinto.core.storage import exceptions as storage_exceptions
from pyramid.settings import asbool
from pyramid_mailer import get_mailer
from pyramid_mailer.message import Message


logger = logging.getLogger(__name__)


EMAIL_REGEXP = re.compile(r"^(.*<[^@<>\s]+@[^@<>\s]+>)|([^@<>\s]+@[^@<>\s]+)$")
GROUP_REGEXP = re.compile(r"^/buckets/[^/]+/groups/[^/]+$")


def qualname(obj):
    """
    >>> str(msg.__class__)
    "<class 'pyramid_mailer.message.Message'>"
    >>> str(msg.__class__).split("'")
    ['<class ', 'pyramid_mailer.message.Message', '>']
    """
    return str(obj.__class__).split("'")[1]


def context_from_event(event):
    root_url = event.request.route_url("hello")
    context = dict(
        event=qualname(event),
        root_url=root_url,
        client_address=event.request.client_addr,
        user_agent=event.request.user_agent,
        **event.payload,
    )

    context["settings"] = event.request.registry.settings

    # The following payload attributes are not always present.
    # See Kinto/kinto#945
    context.setdefault("record_id", "{record_id}")
    context.setdefault("collection_id", "{collection_id}")
    return context


def build_notification(event):
    resource_name = event.payload["resource_name"]
    storage = event.request.registry.storage
    context = context_from_event(event)

    # Build every email for every impacted objects.
    messages = []
    for impacted in event.impacted_objects:
        # Maybe context reliable on batch requests.
        # See Kinto/kinto#945
        _context = context.copy()
        object_id = impacted.get("new", impacted.get("old"))["id"]
        _context[resource_name + "_id"] = _context["id"] = object_id
        messages += get_messages(storage, _context)

    # Store the list in the current request, they will be sent from a
    # post commit hook (we don't send them if DB transaction is rolledback).
    setattr(event.request, "_kinto_emailer_messages", messages)


def send_notification(event):
    # At this point, we can't use `storage` because the transaction was committed.
    messages = event.request._kinto_emailer_messages
    settings = event.request.registry.settings
    mailer = get_mailer(event.request)
    try:
        for message in messages:
            if settings.get("mail.queue_path") is None:
                mailer.send_immediately(message, fail_silently=False)
            else:
                mailer.send_to_queue(message)
    except Exception:
        logger.exception("Could not send notifications")


def _get_emailer_hooks(storage, context):
    bucket_id = context["bucket_id"]
    collection_id = context["collection_id"]
    bucket_uri = "/buckets/%s" % bucket_id
    # Look-up collection metadata.
    metadata = storage.get(
        parent_id=bucket_uri, resource_name="collection", object_id=collection_id
    )
    if "kinto-emailer" not in metadata:
        # Try in bucket metadata.
        metadata = storage.get(parent_id="", resource_name="bucket", object_id=bucket_id)
    # Returns empty list of hooks.
    return metadata.get("kinto-emailer", {}).get("hooks", [])


def _expand_recipients(storage, recipients, context):
    emails = [r for r in recipients if not GROUP_REGEXP.match(r)]
    # Group name using context (eg. /buckets/staging/{collection_id}-reviewers).
    groups = [r.format(**context) for r in recipients if GROUP_REGEXP.match(r.format(**context))]
    # Obtain group members from storage.
    for group_uri in groups:
        bucket_uri, group_id = group_uri.split("/groups/")
        try:
            group = storage.get(parent_id=bucket_uri, resource_name="group", object_id=group_id)
        except storage_exceptions.RecordNotFoundError:
            continue
        # Take out prefix from user ids (e.g. "ldap:mathieu@mozilla.com")
        unprefixed_members = [m.split(":", 1)[-1] for m in group["members"]]
        # Keep only group members that are email addresses.
        emails.extend([m for m in unprefixed_members if EMAIL_REGEXP.match(m)])

    return emails


def _match(hook, context):
    # Allow support of regexps in fields, if they start with ^
    if hook.startswith("^"):
        return re.match(hook, context)
    return hook == context


def get_messages(storage, context):
    hooks = _get_emailer_hooks(storage, context)
    filters = ("event", "action", "resource_name", "id", "record_id", "collection_id")
    messages = []
    for hook in hooks:
        # Filter out hook if it doesn't meet current event attributes, and keep
        # if nothing is specified.
        conditions_met = all(
            [
                field not in hook or field not in context or _match(hook[field], context[field])
                for field in filters
            ]
        )
        if not conditions_met:
            continue

        msg = hook["template"].format(**context)
        subject = hook.get("subject", "New message").format(**context)
        recipients = _expand_recipients(storage, hook["recipients"], context)

        if not recipients:
            continue

        messages.append(
            Message(subject=subject, sender=hook.get("sender"), recipients=recipients, body=msg)
        )
    return messages


def _validate_emailer_settings(event):
    request = event.request
    resource_name = event.payload["resource_name"]

    for impacted in event.impacted_objects:
        # See Kinto/kinto#945
        bid = impacted["new"]["id"] if resource_name == "bucket" else event.payload["bucket_id"]
        bucket_uri = "/buckets/{}".format(bid)

        metadata = impacted["new"]
        if "kinto-emailer" not in metadata:
            continue
        try:
            hooks = metadata["kinto-emailer"]["hooks"]
        except KeyError:
            raise_invalid(request, description='Missing "hooks".')

        for hook in hooks:
            try:
                hook["template"]
            except KeyError:
                raise_invalid(request, description='Missing "template".')

            recipients = hook.get("recipients", [])
            if not recipients:
                raise_invalid(request, description="Empty list of recipients.")

            invalids = [
                r for r in recipients if not (EMAIL_REGEXP.match(r) or GROUP_REGEXP.match(r))
            ]
            if invalids:
                error_msg = "Invalid recipients %s" % ", ".join(invalids)
                raise_invalid(request, description=error_msg)

            invalid_groups = [
                r for r in recipients if GROUP_REGEXP.match(r) and not r.startswith(bucket_uri)
            ]
            if invalid_groups:
                error_msg = "Invalid bucket for groups %s" % ", ".join(invalid_groups)
                raise_invalid(request, description=error_msg)


def includeme(config):
    # Include the mailer
    settings = config.get_settings()
    debug = asbool(settings.get("mail.debug_mailer", "false"))
    config.include("pyramid_mailer" + (".debug" if debug else ""))

    # Expose the capabilities in the root endpoint.
    message = "Provide emailing capabilities to the server."
    docs = "https://github.com/Kinto/kinto-emailer/"
    config.add_api_capability("emailer", message, docs)

    # Listen to collection modification before commit for validation.
    config.add_subscriber(
        _validate_emailer_settings,
        ResourceChanged,
        for_resources=(
            "bucket",
            "collection",
        ),
        for_actions=("create", "update"),
    )

    # Listen to collection and record change events.
    config.add_subscriber(
        build_notification, ResourceChanged, for_resources=("record", "collection")
    )
    config.add_subscriber(
        send_notification, AfterResourceChanged, for_resources=("record", "collection")
    )



import os.path
import shutil
import unittest
from glob import glob
from urllib.parse import urljoin

import requests
from kinto_client import Client


__HERE__ = os.path.abspath(os.path.dirname(__file__))

SERVER_URL = "http://localhost:8888/v1"
DEFAULT_AUTH = ("user", "p4ssw0rd")


class FunctionalTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(FunctionalTest, self).__init__(*args, **kwargs)

        # Setup the kinto clients for the source and destination.
        self._auth = DEFAULT_AUTH
        self._server_url = SERVER_URL
        self._bucket = "emailer"
        self._collection = "collection1"

        self.client = Client(
            server_url=self._server_url,
            auth=self._auth,
            bucket=self._bucket,
            collection=self._collection,
        )
        # Since we use a PG database that can contain objects, start clean.
        self._flush_server(self._server_url)

    def tearDown(self):
        # Delete all the created objects.
        self._flush_server(self._server_url)
        # Delete all files but keep the folder for next runs.
        shutil.rmtree("mail/", ignore_errors=True)
        os.makedirs("mail/")

    def _flush_server(self, server_url):
        flush_url = urljoin(server_url, "/__flush__")
        resp = requests.post(flush_url)
        resp.raise_for_status()

    def test_new_record_send_email(self):
        self.client.create_bucket()
        self.client.create_collection(
            data={
                "kinto-emailer": {
                    "hooks": [
                        {
                            "resource_name": "record",
                            "action": "create",
                            "subject": "Action on {root_url}",
                            "template": (
                                "{client_address} edited:\n "
                                "{bucket_id}/{collection_id}/{record_id}."
                            ),
                            "recipients": ["kinto-emailer@restmail.net"],
                        }
                    ]
                }
            }
        )

        record = self.client.create_record({"foo": "bar"})["data"]

        filename = glob("mail/*.eml")[0]
        with open(filename, "r") as f:
            body = f.read()

        assert "kinto-emailer@restmail.net" in body
        assert "Action on http://localhost:8888/v1/" in body
        assert "127.0.0.1=20edited" in body
        assert "emailer/collection1/{}".format(record["id"]) in body


if __name__ == "__main__":
    unittest.main()



import sys

from pyramid.paster import bootstrap
from pyramid_mailer import get_mailer
from pyramid_mailer.message import Message


subject = "[kinto-emailer] Test"
body = "If you received this email, the server is well configured."


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    try:
        config_file, recipient = args
    except ValueError:
        print("Usage: %s CONFIG RECIPIENT" % sys.argv[0])
        return 1

    print("Load config...")
    env = bootstrap(config_file)

    print("Send email to %r" % recipient)
    registry = env["registry"]
    mailer = get_mailer(registry)

    message = Message(subject=subject, recipients=[recipient], body=body)
    mailer.send_immediately(message, fail_silently=False)
    print("Done.")
    return 0



import configparser
import os
import unittest

import mock
from kinto import main as kinto_main
from kinto.core import errors
from kinto.core.events import AfterResourceChanged
from kinto.core.testing import BaseWebTest, FormattedErrorMixin, get_user_headers

from kinto_emailer import build_notification, context_from_event, get_messages, send_notification


HERE = os.path.dirname(os.path.abspath(__file__))

COLLECTION_RECORD = {
    "kinto-emailer": {
        "hooks": [
            {
                "resource_name": "record",
                "action": "update",
                "sender": "kinto@restmail.net",
                "subject": "Record update",
                "template": "Bonjour les amis.",
                "recipients": ["kinto-emailer@restmail.net"],
            },
            {
                "resource_name": "collection",
                "action": "update",
                "sender": "kinto@restmail.net",
                "subject": "Collection update",
                "template": "Bonjour les amis on collection update.",
                "recipients": ["kinto-emailer@restmail.net"],
            },
        ]
    }
}


class EmailerTest(BaseWebTest, unittest.TestCase):
    entry_point = kinto_main
    api_prefix = "v1"
    config = "config/kinto.ini"

    @classmethod
    def get_app_settings(cls, extras=None):
        ini_path = os.path.join(HERE, cls.config)
        config = configparser.ConfigParser()
        config.read(ini_path)
        settings = dict(config.items("app:main"))
        if extras:
            settings.update(extras)
        return settings


class PluginSetupTest(EmailerTest):
    def test_capability_is_exposed(self):
        resp = self.app.get("/")
        capabilities = resp.json["capabilities"]
        self.assertIn("emailer", capabilities)
        expected = {
            "description": "Provide emailing capabilities to the server.",
            "url": "https://github.com/Kinto/kinto-emailer/",
        }
        self.assertEqual(expected, capabilities["emailer"])

    def test_send_notification_is_called_on_new_record(self):
        with mock.patch("kinto_emailer.send_notification") as mocked:
            app = self.make_app()
            app.post_json(
                "/buckets/default/collections/foobar/records",
                headers={"Authorization": "Basic bmF0aW06"},
            )
            event = mocked.call_args[0][0]
            assert isinstance(event, AfterResourceChanged)

    def test_send_notification_is_called_on_collection_update(self):
        with mock.patch("kinto_emailer.send_notification") as mocked:
            app = self.make_app()
            app.put_json(
                "/buckets/default/collections/foobar",
                {"data": {"status": "update"}},
                headers={"Authorization": "Basic bmF0aW06"},
            )
            event = mocked.call_args[0][0]
            assert isinstance(event, AfterResourceChanged)


class GetMessagesTest(unittest.TestCase):
    def setUp(self):
        self.storage = mock.MagicMock()
        self.storage.get.return_value = COLLECTION_RECORD
        self.payload = {
            "bucket_id": "b",
            "collection_id": "c",
            "resource_name": "record",
            "action": "update",
        }

    def test_get_messages_returns_configured_messages_for_records(self):
        (message,) = get_messages(self.storage, self.payload)

        assert message.subject == "Record update"
        assert message.sender == "kinto@restmail.net"
        assert message.recipients == ["kinto-emailer@restmail.net"]
        assert message.body == "Bonjour les amis."

    def test_get_messages_returns_a_configured_message_for_collection_update(self):
        self.payload.update({"resource_name": "collection"})
        (message,) = get_messages(self.storage, self.payload)

        assert message.subject == "Collection update"
        assert message.sender == "kinto@restmail.net"
        assert message.recipients == ["kinto-emailer@restmail.net"]
        assert message.body == "Bonjour les amis on collection update."

    def test_get_emailer_info_returns_empty_list_if_emailer_not_configured(self):
        self.storage.get.return_value = {}
        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 0

    def test_get_messages_returns_default_subject_to_new_message(self):
        self.storage.get.return_value = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "template": "Bonjour les amis.",
                        "recipients": ["kinto-emailer@restmail.net"],
                    }
                ]
            }
        }
        (message,) = get_messages(self.storage, self.payload)

        assert message.subject == "New message"

    def test_get_messages_returns_several_messages_if_hooks_match(self):
        self.storage.get.return_value = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "template": "Bonjour les amis.",
                        "recipients": ["me@you.com"],
                    },
                    {
                        "template": "Bonjour les amies.",
                        "recipients": ["you@me.com"],
                    },
                ]
            }
        }
        messages = get_messages(self.storage, self.payload)

        assert len(messages) == 2

    def test_get_messages_can_filter_by_id(self):
        self.storage.get.return_value = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "id": "poll",
                        "template": "Poll changed.",
                        "recipients": ["me@you.com"],
                    }
                ]
            }
        }
        self.payload.update({"id": "abc"})
        messages = get_messages(self.storage, self.payload)

        assert len(messages) == 0

    def test_get_messages_can_filter_using_regexps(self):
        self.storage.get.return_value = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "collection_id": "^(?!normandy-recipes$)",
                        "template": "Poll changed.",
                        "recipients": ["me@you.com"],
                    }
                ]
            }
        }
        self.payload.update({"collection_id": "normandy-recipes"})
        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 0

        self.payload.update({"collection_id": "some-normandy-recipes"})
        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 1

        self.payload.update({"collection_id": "normandy-recipes-all"})
        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 1

    def test_get_messages_ignores_resource_if_not_specified(self):
        self.storage.get.return_value = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "template": "Poll changed.",
                        "recipients": ["me@you.com"],
                    }
                ]
            }
        }

        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 1

        self.payload.update({"resource_name": "collection"})
        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 1

    def test_get_messages_ignores_action_if_not_specified(self):
        self.storage.get.return_value = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "template": "Poll changed.",
                        "recipients": ["me@you.com"],
                    }
                ]
            }
        }

        self.payload.update({"action": "create"})
        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 1

        self.payload.update({"action": "update"})
        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 1

    def test_get_messages_can_filter_by_event_class(self):
        self.storage.get.return_value = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "event": "mylib.MyEvent",
                        "template": "Poll changed.",
                        "recipients": ["me@you.com"],
                    }
                ]
            }
        }
        self.payload.update({"event": "kinto.core.events.AfterResourceChanged"})
        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 0

        self.payload.update({"event": "mylib.MyEvent"})
        messages = get_messages(self.storage, self.payload)
        assert len(messages) == 1


class BucketTest(unittest.TestCase):
    def test_hooks_can_be_defined_on_buckets(self):
        storage = mock.MagicMock()
        collection_metadata = {}
        bucket_metadata = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "template": "Poll changed.",
                        "recipients": ["from@bucket.com"],
                    }
                ]
            }
        }
        storage.get.side_effect = [collection_metadata, bucket_metadata]
        payload = {"bucket_id": "b", "collection_id": "c"}
        (message,) = get_messages(storage, payload)
        assert message.recipients == ["from@bucket.com"]


class GroupExpansionTest(unittest.TestCase):
    def setUp(self):
        self.storage = mock.MagicMock()
        self.collection_record = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "template": "Poll changed.",
                        "recipients": ["/buckets/b/groups/g"],
                    }
                ]
            }
        }
        self.group_record = {"members": ["fxa:225689876", "portier:devnull@localhost.com"]}
        self.storage.get.side_effect = [self.collection_record, self.group_record]

    def test_recipients_are_expanded_from_group_members(self):
        payload = {"bucket_id": "b", "collection_id": "c"}
        (message,) = get_messages(self.storage, payload)
        assert message.recipients == ["devnull@localhost.com"]

    def test_no_message_no_email_in_group_members(self):
        group_record = {"members": ["fxa:225689876", "basicauth:toto-la-mamposina"]}
        self.storage.get.side_effect = [self.collection_record, group_record]
        payload = {"bucket_id": "b", "collection_id": "c"}
        messages = get_messages(self.storage, payload)
        assert not messages

    def test_email_group_can_contain_placeholders(self):
        collection_record = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "template": "Poll changed.",
                        "recipients": ["/buckets/b/groups/{collection_id}"],
                    }
                ]
            }
        }
        self.storage.get.side_effect = [collection_record, self.group_record]
        payload = {"bucket_id": "b", "collection_id": "c"}
        get_messages(self.storage, payload)
        self.storage.get.assert_called_with(
            parent_id="/buckets/b", resource_name="group", object_id="c"
        )


class SendNotificationTest(unittest.TestCase):
    def test_send_notification_does_not_call_the_mailer_if_no_message(self):
        event = mock.MagicMock()
        event.payload = {
            "resource_name": "record",
            "action": "update",
            "bucket_id": "default",
            "collection_id": "foobar",
        }
        event.request.registry.storage.get.return_value = {}

        with mock.patch("kinto_emailer.get_mailer") as get_mailer:
            build_notification(event)
            send_notification(event)
            assert not get_mailer().send_immediately.called

    def test_send_notification_calls_the_mailer_if_match_event(self):
        event = mock.MagicMock()
        event.impacted_objects = [{"new": {"id": "a"}}]
        event.payload = {
            "resource_name": "record",
            "action": "update",
            "bucket_id": "default",
            "collection_id": "foobar",
        }
        event.request.registry.storage.get.return_value = COLLECTION_RECORD
        event.request.registry.settings = {}

        with mock.patch("kinto_emailer.get_mailer") as get_mailer:
            build_notification(event)
            send_notification(event)
            assert get_mailer().send_immediately.called

    def test_send_notification_calls_the_mailer_queue_if_configured(self):
        event = mock.MagicMock()
        event.impacted_objects = [{"new": {"id": "a"}}]
        event.payload = {
            "resource_name": "record",
            "action": "update",
            "bucket_id": "default",
            "collection_id": "foobar",
        }
        event.request.registry.storage.get.return_value = COLLECTION_RECORD
        event.request.registry.settings = {"mail.queue_path": "/var/mail"}

        with mock.patch("kinto_emailer.get_mailer") as get_mailer:
            build_notification(event)
            send_notification(event)
            assert get_mailer().send_to_queue.called


class ContextContentTest(unittest.TestCase):
    def test_context_contains_settings(self):
        event = mock.MagicMock()
        event.request.registry.settings = {"project_name": "Kinto DEV"}

        context = context_from_event(event)

        assert "{settings[project_name]}".format(**context) == "Kinto DEV"


class BatchRequestTest(EmailerTest):
    def setUp(self):
        bucket = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "action": "create",
                        "resource_name": "collection",
                        "subject": "Created {bucket_id}/{collection_id}.",
                        "template": "",
                        "recipients": ["me@you.com"],
                    }
                ]
            }
        }
        self.headers = dict(self.headers, **get_user_headers("nous"))
        self.app.put_json("/buckets/b", {"data": bucket}, headers=self.headers)

        patch = mock.patch("kinto_emailer.get_mailer")
        self.get_mailer = patch.start()
        self.addCleanup(patch.stop)

    def test_emails_are_generated_for_each_batch_subrequest(self):
        requests = {
            "defaults": {"method": "POST", "path": "/buckets/b/collections"},
            "requests": [{"body": {"data": {"id": "1"}}}, {"body": {"data": {"id": "2"}}}],
        }
        self.app.post_json("/batch", requests, headers=self.headers)
        call1, call2 = self.get_mailer().send_immediately.call_args_list
        assert call1[0][0].subject == "Created b/1."
        assert call2[0][0].subject == "Created b/2."


class HookValidationTest(FormattedErrorMixin, EmailerTest):
    def setUp(self):
        self.valid_collection = {
            "kinto-emailer": {
                "hooks": [
                    {
                        "template": "{user_id} requested review on {uri}.",
                        "recipients": [
                            "me@you.com",
                            "My friend alice <alice@wonderland.com>",
                            "<t.h.i.s+that@some.crazy.moderndomainnameyouknow>",
                        ],
                    }
                ]
            }
        }
        self.headers = dict(self.headers, **get_user_headers("nous"))
        self.app.put("/buckets/b", headers=self.headers)

    def test_supports_simple_recipients_list(self):
        self.app.put_json(
            "/buckets/b/collections/c", {"data": self.valid_collection}, headers=self.headers
        )

    def test_supports_groups_url(self):
        self.valid_collection["kinto-emailer"]["hooks"][0]["recipients"] += ["/buckets/b/groups/g"]
        self.app.put_json(
            "/buckets/b/collections/c", {"data": self.valid_collection}, headers=self.headers
        )

    def test_supports_empty_list_of_hooks(self):
        self.valid_collection["kinto-emailer"]["hooks"] = []
        self.app.put_json(
            "/buckets/b/collections/c", {"data": self.valid_collection}, headers=self.headers
        )

    def test_supports_empty_template(self):
        self.valid_collection["kinto-emailer"]["hooks"][0]["template"] = ""
        self.app.put_json(
            "/buckets/b/collections/c", {"data": self.valid_collection}, headers=self.headers
        )

    def test_fails_with_missing_hooks(self):
        self.valid_collection["kinto-emailer"].pop("hooks")
        r = self.app.put_json(
            "/buckets/b/collections/c",
            {"data": self.valid_collection},
            headers=self.headers,
            status=400,
        )
        assert 'Missing "hooks"' in r.json["message"]

    def test_bucket_validation_errors_are_well_formatted(self):
        self.valid_collection["kinto-emailer"].pop("hooks")
        r = self.app.post_json(
            "/buckets", {"data": self.valid_collection}, headers=self.headers, status=400
        )
        self.assertFormattedError(
            r,
            400,
            errno=errors.ERRORS.INVALID_PARAMETERS,
            error="Invalid parameters",
            message='Missing "hooks"',
            info=None,
        )

    def test_collection_validation_errors_are_well_formatted(self):
        self.valid_collection["kinto-emailer"].pop("hooks")
        r = self.app.put_json(
            "/buckets/b/collections/c",
            {"data": self.valid_collection},
            headers=self.headers,
            status=400,
        )
        self.assertFormattedError(
            r,
            400,
            errno=errors.ERRORS.INVALID_PARAMETERS,
            error="Invalid parameters",
            message='Missing "hooks"',
            info=None,
        )

    def test_fails_if_missing_template(self):
        self.valid_collection["kinto-emailer"]["hooks"][0].pop("template")
        r = self.app.put_json(
            "/buckets/b/collections/c",
            {"data": self.valid_collection},
            headers=self.headers,
            status=400,
        )
        assert 'Missing "template"' in r.json["message"]

    def test_fails_with_empty_list_of_recipients(self):
        self.valid_collection["kinto-emailer"]["hooks"][0]["recipients"] = []
        r = self.app.put_json(
            "/buckets/b/collections/c",
            {"data": self.valid_collection},
            headers=self.headers,
            status=400,
        )
        assert "Empty list of recipients" in r.json["message"]

    def test_fails_with_invalid_email_recipients(self):
        self.valid_collection["kinto-emailer"]["hooks"][0]["recipients"].extend(
            ["<fe@gmail.com", "haha@haha@com"]
        )
        r = self.app.put_json(
            "/buckets/b/collections/c",
            {"data": self.valid_collection},
            headers=self.headers,
            status=400,
        )
        assert "Invalid recipients <fe@gmail.com, haha@haha@com" in r.json["message"]

    def test_fails_if_group_is_from_other_bucket(self):
        self.valid_collection["kinto-emailer"]["hooks"][0]["recipients"] += [
            "/buckets/plop/groups/g"
        ]
        r = self.app.put_json(
            "/buckets/b/collections/c",
            {"data": self.valid_collection},
            headers=self.headers,
            status=400,
        )
        assert "Invalid bucket for groups /buckets/plop/groups/g" in r.json["message"]

    def test_fails_if_group_uri_is_invalid(self):
        self.valid_collection["kinto-emailer"]["hooks"][0]["recipients"] += [
            "/buckets/b/group/g"  # /groups/g!
        ]
        r = self.app.put_json(
            "/buckets/b/collections/c",
            {"data": self.valid_collection},
            headers=self.headers,
            status=400,
        )
        assert "Invalid recipients /buckets/b/group/g" in r.json["message"]



import unittest

import mock

from kinto_emailer import command_send


class CommandTest(unittest.TestCase):
    def test_uses_sys_args_by_default(self):
        assert command_send.main() > 0  # will fail

    def test_returns_non_zero_if_not_enough_args(self):
        assert command_send.main([]) > 0

    def test_calls_send_immmediately_with_parameters(self):
        with mock.patch("kinto_emailer.command_send.bootstrap"):
            with mock.patch("kinto_emailer.command_send.get_mailer") as get_mailer:
                command_send.main(["config.ini", "me@restmail.net"])

                args, kwargs = get_mailer().send_immediately.call_args_list[0]
                assert "kinto-emailer" in args[0].subject
                assert not kwargs["fail_silently"]



import logging
import sys
import time

import requests
from kinto_http import cli_utils


logger = logging.getLogger(__name__)

DEFAULT_SERVER = "https://remote-settings.allizom.org/v1"
BUCKET = "staging"
COLLECTION = "qa"
MAILBOX = "kintoemailer"
RECIPIENT = MAILBOX + "@restmail.net"
SUBJECT_EMAIL_CREATION = "Record created!"
RETRY_TIMEOUT = 60


def clear_inbox(mailbox):
    requests.delete("https://restmail.net/mail/%s/" % mailbox)


def fetch_emails(mailbox):
    resp = requests.get("https://restmail.net/mail/%s/" % mailbox)
    emails = resp.json()
    return emails


def setup_notifs_signing(client, recipient):
    setup = {
        "kinto-emailer": {
            "hooks": [
                {
                    "event": "kinto_remote_settings.signer.events.ReviewRequested",
                    "subject": "{user_id} requested review on {bucket_id}/{collection_id}.",
                    "template": "Review changes at {root_url}admin/#/buckets/{bucket_id}/collections/{collection_id}/records",
                    "recipients": [recipient],
                },
                {
                    "resource_name": "record",
                    "action": "create",
                    "subject": SUBJECT_EMAIL_CREATION,
                    "template": "For QA purposes :)",
                    "recipients": [recipient],
                },
            ]
        }
    }
    client.patch_collection(data=setup)


def main(args=None):
    parser = cli_utils.add_parser_options(
        description="Validate kinto-emailer setup",
        default_server=DEFAULT_SERVER,
        default_bucket=BUCKET,
        default_collection=COLLECTION,
    )

    args = parser.parse_args(args)

    cli_utils.setup_logger(logger, args)

    client = cli_utils.create_client_from_args(args)

    # 1. Check capabilities
    capabilities = client.server_info()["capabilities"]
    if "emailer" not in capabilities:
        print("Server doesn't have support for email notifications: \x1b[1;31m KO \x1b[0m")
        return 2

    print("Clear inbox")
    clear_inbox(MAILBOX)

    # 2. Configure collection metadata.
    print("Configure emailing in %r collection" % args.collection)
    setup_notifs_signing(client, RECIPIENT)

    # 3. Create a record.
    print("Create a dummy record")
    client.create_record(data={"script": "validate kinto-emailer setup"})

    # 4. If signing is enabled for this collection, then also check email for ReviewRequested.
    signed_resources = capabilities.get("signer", {"resources": []})["resources"]
    ids = [(r["source"]["bucket"], r["source"]["collection"]) for r in signed_resources]
    signing_enabled = (args.bucket, args.collection) in ids

    if signing_enabled:
        # 4. Request a review.
        client.patch_collection(data={"status": "to-review"})

    # 5. Check that emails were received.
    print("Scan inbox...")
    first_try = time.time()
    while True:
        emails = fetch_emails(MAILBOX)
        subjects = [e["subject"] for e in emails]

        if SUBJECT_EMAIL_CREATION not in subjects:
            if time.time() - first_try < RETRY_TIMEOUT:
                time.sleep(3)
                sys.stdout.write(".")
                sys.stdout.flush()
                continue
        break

    if SUBJECT_EMAIL_CREATION in subjects:
        print("Email received:\x1b[1;32m OK \x1b[0m")

        if signing_enabled:
            if "requested review" in subjects:
                print("Review notification received:\x1b[1;32m OK \x1b[0m")
            else:
                print("Review notification not received:\x1b[1;31m KO \x1b[0m")
    else:
        print("Email not received:\x1b[1;31m KO \x1b[0m")

    return 0


if __name__ == "__main__":
    sys.exit(main())
